;; Generated package description from qrencode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "qrencode" "1.4" "QRCode encoder" '((emacs "25.1")) :commit "677939b0566295e1ff227de115edfcfeceb52bab" :authors '(("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net")) :maintainer '("Rüdiger Sonderfeld" . "ruediger@c-plusplus.net") :keywords '("qrcode" "comm") :url "https://github.com/ruediger/qrencode-el")
